import aamp
from functools import lru_cache
from pathlib import Path, WindowsPath, PosixPath
import json
from os import path
from typing import Union
from aamp.parameters import ParameterIO, ParameterList, ParameterObject
from aamp.botw_hashed_names import hash_to_name_map


EXEC_DIR = path.dirname(path.realpath(__file__))

def get_ai_defs() -> dict:
    if not hasattr(get_ai_defs, 'defs'):
        with open(path.join(EXEC_DIR, 'resources', 'aidef.json'), 'r', encoding='utf-8') as file:
            get_ai_defs.defs = json.load(file)
    return get_ai_defs.defs


def get_trans_map() -> dict:
    if not hasattr(get_trans_map, 'map'):
        with open(path.join(EXEC_DIR, 'resources', 'jpen.json'), 'r', encoding='utf-8') as file:
            get_trans_map.map = json.load(file)
    return get_trans_map.map


def get_ai_classes() -> list:
    classes = []
    defs = get_ai_defs()
    for key in defs:
        classes.extend(list(defs[key].keys()))
    return classes


def cjk_detect(texts):
    import re
    if re.search("[\uac00-\ud7a3]", texts) or re.search("[\u3040-\u30ff]", texts) or re.search("[\u4e00-\u9FFF]", texts):
        return True
    return False


def json_serialize(f) -> str:
    def serializer(self, **kwargs):
        data = f(self, **kwargs)
        try:
            json_data = self._encoder.default(data)
            return json_data
        except TypeError:
            return data
    return serializer


def add_hashes():
    if not hasattr(add_hashes, 'done'):
        with open(path.join(EXEC_DIR, 'resources', 'hashes.json'), 'r', encoding='utf-8') as h_file:
            new_hashes = { int(k): v for k,v in json.load(h_file).items() }
        hash_to_name_map.update(new_hashes)
        add_hashes.done = True


def _try_name(key: int) -> Union[str, int]:
    add_hashes()
    tmap = get_trans_map()
    if key in hash_to_name_map:
        s_key = hash_to_name_map[key]
        return s_key
    else:
        return _try_numbered_names(key)

def _try_numbered_names(key: int) -> Union[str, int]:
    from zlib import crc32
    for i in range(1000):
        if crc32(f'AI_{i}'.encode()) == key:
            return f'AI_{i}'
        elif crc32(f'Action_{i}'.encode()) == key:
            return f'Action_{i}'
        elif crc32(f'Behavior_{i}'.encode()) == key:
            return f'Behavior_{i}'
        elif crc32(f'Query_{i}'.encode()) == key:
            return f'Query_{i}'
    return key


class AiProgJsonEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, ParameterIO):
            return self.encode_pio(o)
        elif isinstance(o, ParameterList):
            return self.encode_plist(o)
        elif isinstance(o, ParameterObject):
            return self.encode_pobject(o)
        elif isinstance(o, (Path, WindowsPath, PosixPath)):
            return str(o)
        else:
            return super().default(o)

    def encode_pio(self, pio: ParameterIO) -> dict:
        return {
            'type': pio.type,
            'version': pio.version,
            'lists': { _try_name(k): self.encode_plist(pl) for k, pl in pio.lists.items() },
            'objects': { _try_name(k): self.encode_pobject(pobj) for k, pobj in pio.objects.items() },
        }

    def encode_plist(self, plist: ParameterList) -> dict:
        lists = { _try_name(k): self.encode_plist(pl) for k, pl in plist.lists.items() }
        return {
            'lists': lists,
            'objects': { _try_name(k): self.encode_pobject(pobj) for k, pobj in plist.objects.items() },
        }

    def encode_pobject(self, obj: ParameterObject) -> dict:
        return {
            'params': { _try_name(k): self.encode_param(param) for k, param in obj.params.items() }
        }

    def encode_param(self, param) -> dict:
        t = type(param).__name__
        encoded = None
        if isinstance(param, aamp.Vec2):
            encoded = self._encode_vec2(param)
        elif isinstance(param, aamp.Vec3):
            encoded = self._encode_vec3(param)
        elif isinstance(param, aamp.Vec4):
            encoded = self._encode_vec4(param)
        elif isinstance(param, aamp.Color):
            encoded = self._encode_color(param)
        elif isinstance(param, aamp.Quat):
            encoded = self._encode_quat(param)
        elif isinstance(param, aamp.Curve):
            encoded = self._encode_curve(param)
        elif type(param) in [aamp.String32, aamp.String64, aamp.String256]:
            encoded = self._encode_str(param)
            if encoded in get_trans_map():
                encoded = get_trans_map()[encoded]
        elif isinstance(param, aamp.U32):
            encoded = self._encode_u32(param)
        else:
            encoded = param
        return { t: encoded }

    def _encode_vec2(self, vec: aamp.Vec2) -> tuple:
        return vec.y, vec.z

    def _encode_vec3(self, vec: aamp.Vec3) -> tuple:
        return vec.x, vec.y, vec.z

    def _encode_vec4(self, vec: aamp.Vec4) -> tuple:
        return vec.w, vec.x, vec.y, vec.z

    def _encode_color(self, color: aamp.Color) -> tuple:
        return {
            'a': color.a,
            'r': color.r,
            'g': color.g,
            'b': color.b
        }

    def _encode_quat(self, quat: aamp.Quat) -> tuple:
        return quat.a, quat.b, quat.c, quat.d

    def _encode_curve(self, curve: aamp.Curve) -> tuple:
        return tuple(*curve.v)

    def _encode_str(self, string: Union[aamp.String32, aamp.String64, aamp.String256]) -> str:
        return str(string)

    def _encode_u32(self, u32: aamp.U32) -> int:
        return int(u32)


class AiProgJsonDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        try:
            iter(obj)
        except:
            return obj
        if isinstance(obj, ParameterList) or isinstance(obj, ParameterIO)\
           or isinstance(obj, ParameterObject) or isinstance (obj, str) or isinstance (obj, int)\
           or isinstance(obj, aamp.String32):
            return obj
        if 'version' in obj:
            return self._to_pio(obj)
        elif 'lists' in obj:
            return self._to_plist(obj)
        elif 'params' in obj:
            return self._to_pobj(obj)
        elif 'int' in obj:
            return int(obj['int'])
        elif 'str' in obj:
            return str(obj['str'])
        elif 'float' in obj:
            return float(obj['float'])
        elif 'String32' in obj:
            return aamp.String32(obj['String32'])
        elif 'bool' in obj:
            return obj['bool']
        elif 'Vec3' in obj:
            return aamp.Vec3(*obj['Vec3'])
        else:
            return obj

    def _to_pobj(self, obj) -> aamp.ParameterObject:
        if isinstance(obj, ParameterObject):
            return obj
        pobj = aamp.ParameterObject()
        if obj['params']:
            for param, val in obj['params'].items():
                if param.isnumeric():
                    pobj.params[int(param)] = self.object_hook(val)
                else:
                    pobj.set_param(param, self.object_hook(val))
        return pobj

    def _to_plist(self, obj) -> aamp.ParameterList:
        plist = aamp.ParameterList()
        if isinstance(obj, ParameterList):
            return obj
        if obj['lists']:
            for name, content in obj['lists'].items():
                if name.isnumeric():
                    plist.lists[int(name)] = self._to_plist(content)
                else:
                    plist.set_list(name, self._to_plist(content))
        if obj['objects']:
            for name, content in obj['objects'].items():
                if content['params']:
                    if name.isnumeric():
                        plist.objects[int(name)] = self._to_pobj(content)
                    else:
                        plist.set_object(name, self._to_pobj(content))
        return plist

    def _to_pio(self, obj) -> aamp.ParameterIO:
        pio = aamp.ParameterIO(obj['type'], obj['version'])
        if obj['lists']:
            for name, content in obj['lists'].items():
                if name.isnumeric():
                    pio.lists[int(name)] = self._to_plist(content)
                else:
                    pio.set_list(name, self._to_plist(content))
        if obj['objects']:
            for name, content in obj['objects'].items():
                if content['params']:
                    if name.isnumeric():
                        pio.objects[int(name)] = self._to_pobj(content)
                    else:
                        pio.set_object(name, self._to_pobj(content))
        return pio



from copy import deepcopy
import json
from typing import List, Union
from zlib import crc32
import aamp
from aamp.parameters import ParameterIO, ParameterList, ParameterObject, ParameterType
from . import util


class AiProgram:

    _aiprog: ParameterIO
    _ais: List[ParameterList]
    _actions: List[ParameterList]
    _behaviors: List[ParameterList]
    _queries: List[ParameterList]

    def __init__(self, data: ParameterIO):
        self._aiprog = data
        total = 0
        self._ais = list(data.list('param_root').list('AI').lists.values())
        total += len(self._ais)
        self._actions = list(data.list('param_root').list('Action').lists.values())
        total += len(self._actions)
        self._behaviors = list(data.list('param_root').list('Behavior').lists.values())
        total += len(self._behaviors)
        self._queries = list(data.list('param_root').list('Query').lists.values())

    def get_references(self, idx: int) -> {}:
        idxs = {
            'demo': [],
            'ai': {
                'child': {},
                'behavior': {}
            },
            'action': {
                'behavior': {}
            }
        }
        for key, i in self._aiprog.list('param_root').object('DemoAIActionIdx').params.items():
            if i == idx:
                idxs['demo'].append(key)
        for i, ai in enumerate(self._ais):
            for key, index in ai.object('ChildIdx').params.items():
                if index == idx:
                    if not i in idxs['ai']['child']:
                        idxs['ai']['child'][i] = []
                    idxs['ai']['child'][i].append(key)
            if crc32('BehaviorIdx'.encode()) in ai.objects:
                for key, index in ai.object('BehaviorIdx').params.items():
                    if index == idx:
                        if not i in idxs['ai']['behavior']:
                            idxs['ai']['behavior'][i] = []
                        idxs['ai']['behavior'][i].append(key)
        for i, action in enumerate(self._actions):
            if crc32('BehaviorIdx'.encode()) in action.objects:
                for key, index in action.object('BehaviorIdx').params.items():
                    if index == idx:
                        if not i in idxs['action']['behavior']:
                            idxs['action']['behavior'][i] = []
                        idxs['action']['behavior'][i].append(key)
        return idxs

    def get_actions_offset(self) -> int:
        return len(self._ais)

    def get_behaviors_offset(self) -> int:
        return len(self._ais) + len(self._actions)
    
    def get_queries_offset(self) -> int:
        return len(self._ais) + len(self._actions) + len(self._behaviors)

    def get_ai_at_index(self, idx: int) -> ParameterList:
        return self._ais[idx]

    def get_action_at_index(self, idx: int) -> ParameterList:
        return self._actions[idx - self.get_actions_offset()]

    def get_behavior_at_index(self, idx: int, behavior_idx: bool = False) -> ParameterList:
        if not behavior_idx:
            return self._behaviors[idx - self.get_behaviors_offset()]
        else:
            return self._behaviors[idx]

    def get_query_at_index(self, idx: int) -> ParameterList:
        return self._queries[idx - self.get_queries_offset()]

    def get_item_at_index(self, idx: int) -> (str, ParameterList):
        if idx < self.get_actions_offset():
            return ('ai', self.get_ai_at_index(idx))
        elif self.get_actions_offset() <= idx < self.get_behaviors_offset():
            return ('action', self.get_action_at_index(idx))
        elif self.get_behaviors_offset() <= idx < self.get_queries_offset():
            return ('behavior', self.get_behavior_at_index(idx))
        else:
            return ('query', self.get_query_at_index(idx))

    def get_behavioridx(self, item: ParameterList) -> int:
        return self._behaviors.index(item)

    def items(self) -> tuple:
        return (*self._ais, *self._actions, *self._behaviors, *self._queries)

    def index(self, item: ParameterList) -> int:
        return self.items().index(item)

    def index_from_label(self, label: str) -> int:
        if label.startswith('AI'):
            return int(label.replace('AI_', ''))
        elif label.startswith('Action'):
            return int(label.replace('Action_', '')) + self.get_actions_offset()
        elif label.startswith('Behavior'):
            return int(label.replace('Behavior_', '')) + self.get_behaviors_offset()
        elif label.startswith('Query'):
            return int(label.replace('Query_', '')) + self.get_queries_offset()
        else:
            raise ValueError(f'{label} is not a valid AI entry.')

    def _remove_refs(self, idx: int):
        self._update_indexes(idx, 0)

    def _update_indexes(self, old_idx: int, new_idx: int):
        refs = self.get_references(old_idx)
        for param in refs['demo']:
            self._aiprog.list('param_root').object('DemoAIActionIdx').params[param] = new_idx
        for ai_idx, child_list in refs['ai']['child'].items():
            for child in child_list:
                self._ais[ai_idx].object('ChildIdx').params[child] = new_idx
        for be_idx, be_list in refs['ai']['behavior'].items():
            for be in be_list:
                self._ais[be_idx].object('BehaviorIdx').params[be] = \
                    new_idx - self.get_behaviors_offset()
        for be_idx, be_list in refs['action']['behavior'].items():
            for be in be_list:
                self._actions[be_idx].object('BehaviorIdx').params[be] = \
                    new_idx - self.get_behaviors_offset()             

    def add_ai(self, item: ParameterList) -> int:
        item = deepcopy(item)
        for thing in reversed([*self._actions, *self._behaviors, *self._queries]):
            self._update_indexes(self.index(thing), self.index(thing) + 1)
        if crc32(b'ChildIdx') in item.objects:
            for param in item.object('ChildIdx').params:
                item.object('ChildIdx').params[param] = -1
        if crc32(b'BehaviorIdx') in item.objects:
            for param in item.object('BehaviorIdx').params:
                item.object('BehaviorIdx').params[param] = -1
        self._ais.append(item)
        return self.index(item)

    def add_action(self, item: ParameterList) -> int:
        item = deepcopy(item)
        for thing in [*self._behaviors, *self._queries]:
            self._update_indexes(self.index(thing), self.index(thing) + 1)
        if crc32(b'BehaviorIdx') in item.objects:
            for param in item.object('BehaviorIdx').params:
                item.object('BehaviorIdx').params[param] = 0
        self._actions.append(item)
        return self.index(item)

    def add_behavior(self, item: ParameterList) -> int:
        item = deepcopy(item)
        for query in self._queries:
            self._update_indexes(self.index(query), self.index(query) + 1)
        self._behaviors.append(item)
        return self.index(item)

    def add_query(self, item: ParameterList) -> int:
        item = deepcopy(item)
        self._queries.append(item)
        return self.index(item)

    def remove_item(self, item: Union[int, ParameterList]) -> ParameterList:
        if isinstance(item, int):
            idx = item
            item = self.get_item_at_index(idx)
        else:
            idx = self.index(item)
        for i in [i for i, stuff in enumerate(self.items()) if i >= idx]:
            self._update_indexes(i, i - 1)
        if item[0] == 'ai':
            self._ais.remove(item[1])
        elif item[0] == 'action':
            self._actions.remove(item[1])
        elif item[0] == 'behavior':
            self._behaviors.remove(item[1])
        else:
            self._queries.remove(item)

    def generate_pio(self) -> aamp.ParameterIO:
        from zlib import crc32
        pio = deepcopy(self._aiprog)
        if self._ais:
            ais_list = aamp.ParameterList()
            for idx, ai in enumerate(self._ais):
                ais_list.set_list(f'AI_{idx}', ai)
            pio.list('param_root').set_list('AI', ais_list)
        if self._actions:
            actions_list = aamp.ParameterList()
            for idx, action in enumerate(self._actions):
                actions_list.set_list(f'Action_{idx}', action)
            pio.list('param_root').set_list('Action', actions_list)
        if self._behaviors:
            behaviors_list = aamp.ParameterList()
            for idx, behavior in enumerate(self._behaviors):
                behaviors_list.set_list(f'Behavior_{idx}', behavior)
            pio.list('param_root').set_list('Behavior', behaviors_list)
        if self._queries:
            queries_list = aamp.ParameterList()
            for idx, query in enumerate(self._queries):
                queries_list.set_list(f'Query_{idx}', query)
            pio.list('param_root').set_list('Query', queries_list)
        self._aiprog = pio
        return pio

    def get_roots(self) -> set:
        roots = set()
        for i, ai_item in enumerate(self._ais):
            refs = self.get_references(i)
            if not refs['ai']['child']:
                roots.add(i)
        return roots

    def get_tree(self) -> []:
        trees = []
        for r in self.get_roots():
            trees.append(self._generate_tree_node(r))
        return trees

    def _generate_tree_node(self, idx) -> dict:
        itms = self.items()
        jpen = util.get_trans_map()
        ai: ParameterList = itms[idx]
        try:
            text = util._try_name(ai.object('Def').param('Name'))
        except KeyError:
            text = util._try_name(str(ai.object('Def').param('ClassName')))
        tree = {
            'text': text if not text in jpen else jpen[text],
            'expanded': True,
            'index': idx
        }
        try:
            for child, i in ai.object('ChildIdx').params.items():
                if 'nodes' not in tree:
                    tree['nodes'] = []
                if i >= 0:
                    tree['nodes'].append(self._generate_tree_node(i))
        except KeyError:
            pass
        return tree

_param_type_map = {
    'String': aamp.String32,
    'Bool': bool,
    'Int': int,
    'Float': float,
    'Vec3': aamp.Vec3
}

class ProgramAI(ParameterList):

    def __init__(self, def_name: str, type: str, name: str = '', group: str = ''):
        super().__init__()
        defs = util.get_ai_defs()
        key = 'AIs' if type == 'ai' else 'Actions' if type == 'action'\
              else 'Behaviors' if type == 'behavior' else 'Querys' if type == 'query' else ''
        if def_name in defs[key]:
            ai_def = defs[key][def_name]
            self.lists = {}
            self.objects = {
                crc32(b'Def'): ParameterObject()
            }
            self.object('Def').set_param('ClassName', aamp.String32(def_name))
            self.object('Def').set_param('Name', name)
            self.object('Def').set_param('GroupName', group)
            if 'childs' in ai_def:
                self.set_object('ChildIdx', ParameterObject())
                for child in ai_def['childs']:
                    self.object('ChildIdx').set_param(child, -1)
            if 'StaticInstParams' in ai_def:
                self.set_object('SInst', ParameterObject())
                for sinst in ai_def['StaticInstParams']:
                    if 'Value' in sinst:
                        val = sinst['Value']
                    else:
                        val = _param_type_map[sinst['Type']]()
                    self.object('SInst').set_param(sinst['Name'], val)
        else:
            raise ValueError(f'{def_name} is not a valid AI')




from importlib.util import find_spec
import json
import os
from pathlib import Path
from platform import system
import aamp
import webview
from . import DEBUG
from . import ai, util

class Api:
    def __init__(self):
        self._encoder = util.AiProgJsonEncoder()
        self._decoder = util.AiProgJsonDecoder()

    @util.json_serialize
    def init(self):
        return { 'defs': util.get_ai_defs(), 'trans': util.get_trans_map() }

    @util.json_serialize
    def open_file(self):
        result = self.window.create_file_dialog(webview.OPEN_DIALOG, file_types=(
            'AI Programs (*.baiprog)', 'All Files (*.*)'
        ))
        if result:
            open_path = Path(result[0])
            if open_path.exists():
                pio = aamp.Reader(open_path.read_bytes()).parse()
                res = {
                    'path': open_path.as_posix(),
                    'pio': self._encoder.default(pio)
                }
                return res
        else:
            return {
                'path': ''
            }

    def save_file(self, params):
        if 'path' in params and params['path']:
            open_path = Path(params['path'])
        else:
            result = self.window.create_file_dialog(webview.SAVE_DIALOG, file_types=(
                'AI Programs (*.baiprog)', 'All Files (*.*)'
            ))
            if result:
                open_path = Path(result)
            else:
                return { 'success': False }
        try:
            pio = self._decoder.object_hook(params['pio'])
            open_path.write_bytes(aamp.Writer(pio).get_bytes())
            return { 'success': True }
        except Exception as e:
            return { 'success': False, 'error': str(e) }

    @util.json_serialize
    def create_prog_item(self, params):
        return ai.ProgramAI(params['def'], params['type'])

    def get_ai_tree(self, params):
        tree = ai.AiProgram(self._decoder.object_hook(params['pio'])).get_tree()
        return tree

    @util.json_serialize
    def add_ai_item(self, params):
        ai_item = self._decoder.object_hook(params['ai'])
        new_prog = ai.AiProgram(self._decoder.object_hook(params['prog']['pio']))
        if params['type'] == 'ai':
            new_prog.add_ai(ai_item)
        elif params['type'] == 'action':
            new_prog.add_action(ai_item)
        elif params['type'] == 'behavior':
            new_prog.add_behavior(ai_item)
        elif params['type'] == 'query':
            new_prog.add_query(ai_item)
        return new_prog.generate_pio()

    @util.json_serialize
    def remove_ai_item(self, params):
        new_prog = ai.AiProgram(self._decoder.object_hook(params['prog']['pio']))
        new_prog.remove_item(new_prog.index_from_label(params['idx']))
        return new_prog.generate_pio()


def main():
    api = Api()
    api.window = webview.create_window('Plasticity', url=os.path.join(util.EXEC_DIR, 'assets', 'index.html'), js_api=api, text_select=DEBUG, width=1024)
    use_cef = find_spec('cefpython3') is not None
    gui: str = ''
    if system() == 'Windows' and use_cef:
        gui = 'cef'
    elif system() == 'Linux':
        gui = 'qt'
    webview.start(args=api.window, gui=gui, debug=DEBUG)

if __name__ == "__main__":
    main()



from setuptools import setup
from plasticity_ai import VERSION

with open('docs/README.md', 'r') as readme:
    long_description = readme.read()

setup(
    name='plasticity-ai',
    version=VERSION,
    author='NiceneNerd',
    author_email='macadamiadaze@gmail.com',
    description='An AI program editor for The Legend of Zelda: Breath of the Wild',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/NiceneNerd/Plasticity/',
    include_package_data = True,
    packages = ['plasticity_ai'],
    entry_points = {
        'gui_scripts': [
            'plasticity=plasticity_ai.__main__:main'
        ]
    },
    classifiers = [
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.7'
    ],
    python_requires = '>=3.7',
    install_requires = [
        'aamp',
        'pywebview'
    ],
    extras_require={
        'CEF':  ["cefpython3>=66.0"]
    }
)